#include <stdio.h>
#include <stdlib.h>
#include <libxml/xmlreader.h>
#include <string.h>

#include "line.h"

int line_by_line;

void check_line(char *s, char *title, long pos)
{
  char *cur = s;
  char *start = 0;

  while (*cur) {
    if (*cur == '[' && *(cur+1) == '[') {
      start = cur;
      cur += 2;
      while (*cur && (*cur != ']' || *(cur+1) != ']')) cur++;
      if (!*cur) goto error;
      cur += 2;
      continue;
    }
    if (*cur == '{' && *(cur+1) == '{') {
      start = cur;
      while (*cur && (*cur != '}' || *(cur+1) != '}')) cur++;
      if (!*cur) goto error;
      cur += 2;
      continue;
    }
    cur++;
  }

  return;

error:
  if (line_by_line)
    printf("error with '%s' (title %s) (pos %ld)\n", s, title, pos);
  else
    printf("error with '%s' (title %s) (pos %ld)\n", start, title, pos);
}

void do_word(xmlTextReaderPtr reader)
{
  char *text = NULL;
  char *title = NULL;
  int type;
  const char *name;
  int ns=-1;
  int ret;
  line l;
  long text_pos=0;

  while (1) {
    ret = xmlTextReaderRead(reader);
    if (!ret) break;
    type = xmlTextReaderNodeType(reader);
    name = (const char *)xmlTextReaderConstName(reader);
    if (type == XML_READER_TYPE_END_ELEMENT && !strcmp(name, "page")) break;
    if (type == XML_READER_TYPE_ELEMENT && !strcmp(name, "ns")) {
      const char *s = (const char *)xmlTextReaderReadString(reader);
      if (!s) abort();
      if (sscanf(s, "%d", &ns) != 1) abort();
      free((char *)s);
    }
    if (type == XML_READER_TYPE_ELEMENT && !strcmp(name, "text")) {
      if (text) abort();
      text_pos = xmlTextReaderByteConsumed(reader);
      text = (char *)xmlTextReaderReadString(reader);
      if (!text) abort();
    }
    if (type == XML_READER_TYPE_ELEMENT && !strcmp(name, "title")) {
      if (title) abort();
      title = (char *)xmlTextReaderReadString(reader);
      if (!title) abort();
    }
  }

  if (ns != 0) goto end;

  if (!title || !text) abort();

  if (line_by_line) {
    /* work line by line */
    start_line(&l, text);
    while (1) {
      char *nl = next_line(&l);
      if (!nl) break;
      check_line(nl, title, text_pos);
    }
    end_line(&l);
  } else
    /* work on full text (much less false positive, may miss some...) */
    check_line(text, title, text_pos);

end:
  free(title);
  free(text);
}

void do_words(const char *f)
{
  xmlTextReaderPtr reader;
  int ret;
  reader = xmlNewTextReaderFilename(f); if (!reader) abort();
  ret = xmlTextReaderRead(reader);
  while (ret == 1) {
    const unsigned char *name = xmlTextReaderConstName(reader);
    if (xmlTextReaderNodeType(reader) == XML_READER_TYPE_ELEMENT &&
        !strcmp((const char *)name, "page"))
      do_word(reader);
    ret = xmlTextReaderRead(reader);
  }
  xmlFreeTextReader(reader);
}

void usage(void)
{
  printf("usage: check <options> <file>\n");
  printf("options:\n");
  printf("\t-l\t\twork line by line (many false positives)\n");
  exit(0);
}

int main(int n, char **v)
{
  const char *file = NULL;
  int i;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-l")) line_by_line = 1;
    else { if (file) usage(); file = v[i]; }
  }

  if (file == NULL) usage();

  do_words(file);
  return 0;
}
